//	TWI(I2C){색Cui}X^[pj
//	wt]xH쎺	http://jsdiy.web.fc2.com
/*
yXVz
2013/09/27	v1.00	
*/

#include <avr/io.h>
#include "twilib.h"

enum E_TWI_StatusCode
{
	STM_START		= 0x08,	//A START condition has been transmitted
	STM_REPEAT		= 0x10,	//A repeated START condition has been transmitted
	
	//Master Transmitter Mode
	STMT_SLAW_ACK	= 0x18,	//SLA+W has been transmitted; ACK has been received
	STMT_SLAW_NACK	= 0x20,	//SLA+W has been transmitted; NOT ACK has been received
	STMT_DATA_ACK	= 0x28,	//Data byte has been transmitted; ACK has been received
	STMT_DATA_NACK	= 0x30,	//Data byte has been transmitted; NOT ACK has been received
	STMT_ARBI_LOST	= 0x38,	//Arbitration lost in SLA+W or data bytes
	
	//Master Receiver Mode
	STMR_ARBI_LOST	= 0x38,	//Arbitration lost in SLA+R or NOT ACK bit
	STMR_SLAR_ACK	= 0x40,	//SLA+R has been transmitted; ACK has been received
	STMR_SLAR_NACK	= 0x48,	//SLA+R has been transmitted; NOT ACK has been received
	STMR_DATA_ACK	= 0x50,	//Data byte has been received; ACK has been returned
	STMR_DATA_NACK	= 0x58,	//Data byte has been received; NOT ACK has been returned
	
	//Miscellaneous States
	STM_NOTHING		= 0xF8,	//No relevant state information available; TWINT = "0"
	STM_BUS_ERROR	= 0x00	//Bus error due to an illegal START or STOP condition
	
	/*
	//Slave Receiver Mode
	STSR_SLAW_ACK	= 0x60,	//Own SLA+W has been received; ACK has been returned
	STSR_ARBI_LOST_SLAW_ACK	= 0x68,	//Arbitration lost in SLA+R/W as Master; own SLA+W has been received; ACK has been returned
	STSR_GENERALCALL_ACK	= 0x70,	//General call address has been received; ACK has been returned
	STSR_ARBI_LOST_GENERALCALL_ACK	= 0x78,	//Arbitration lost in SLA+R/W as Master; General call address has been received; ACK has been returned
	STSR_PREV_ADDR_SLAW_ACK	= 0x80,	//Previously addressed with own SLA+W; data has been received; ACK has been returned
	STSR_PREV_ADDR_SLAW_NACK	= 0x88,	//Previously addressed with own SLA+W; data has been received; NOT ACK has been returned
	STSR_PREV_ADDR_GENERALCALL_ACK	= 0x90,	//Previously addressed with general call; data has been received; ACK has been returned
	STSR_PREV_ADDR_GENERALCALL_NACK	= 0x98,	//Previously addressed with general call; data has been received; NOT ACK has been returned
	STSR_STOP	= 0xA0,	//A STOP condition or repeated START condition has been received while still addressed as Slave
	
	//Slave Transmitter Mode
	STST_SLAR	= 0xA8,	//Own SLA+R has been received;
	STST_ARBI_LOST_SLAR_ACK	= 0xB0,	// Arbitration lost in SLA+R/W as Master; own SLA+R has been received; ACK has been returned
	STST_DATA_ACK	= 0xB8,	//Data byte in TWDR has been transmitted; ACK has been received
	STST_DATA_NACK	= 0xC0,	//Data byte in TWDR has been transmitted; NOT ACK has been received
	STST_LASTDATA_ACK	= 0xC8	//Last data byte in TWDR has been transmitted (TWEA = g0h); ACK has been received
	*/
};

#define Status	(TWSR & 0xF8)
#define SLA_W_BIT	0
#define SLA_R_BIT	1

//X^[g𑗐M
//߂l	0:, 0:s
uint8_t	TWI_Start()
{
	TWCR = (1 << TWINT) | (1 << TWSTA) | (1 << TWEN);
	while (!(TWCR & (1 << TWINT)));
	return (Status == STM_START || Status == STM_REPEAT);
}

//oXG[mFƑΉ
static	void	BusErrorCheck()
{
	//oXG[ TWCR:TWINT=1, TWCR:TWSTO=1 ƂiXgbv𑗐M΂悢j
	if (Status == STM_BUS_ERROR) TWI_Stop();
}

//X[ufoCX̃AhX^Cӂ̃f[^𑗐M
//߂l:	Xe[^X
static	uint8_t	MasterTransmit(uint8_t byteValue)
{
	TWDR = byteValue;
	TWCR = (1 << TWINT) | (1 << TWEN);
	while (!(TWCR & (1 << TWINT)));
	BusErrorCheck();
	return Status;
}

//X[ufoCX̃AhXw肷Fݕ
//:	(AhX7bit){(1bit)ō\ꂽ8bit̒l
//߂l	0:, 0:s
uint8_t	TWI_SlaveAddrW(uint8_t adrs)
{
	return (MasterTransmit(adrs | SLA_W_BIT) == STMT_SLAW_ACK);
}

//X[ufoCX̃AhXw肷Fǂݍݕ
//:	(AhX7bit){(1bit)ō\ꂽ8bit̒l
//߂l	0:, 0:s
uint8_t	TWI_SlaveAddrR(uint8_t adrs)
{
	return (MasterTransmit(adrs | SLA_R_BIT) == STMR_SLAR_ACK);
}

//f[^𑗐M
//߂l	0:, 0:s
uint8_t	TWI_WriteByte(uint8_t data)
{
	return (MasterTransmit(data) == STMT_DATA_ACK);
}

//f[^M
//	retAck:	1:ACKԂ, 0:NACKԂ
//		data: ǂݍ񂾒li[
//߂l	0:, 0:s
//
//yFI2C-EEPROM̃V[PVǂݍ݁z
//E ACKԂc f[^1MAĎ̃f[^vꍇB
//ENACKԂc f[^1MĊꍇiɃXgbv𑗐MjB
uint8_t	TWI_ReadByte(uint8_t retAck, uint8_t* data)
{
	//X[u̓߂
	//Ẽf[^M̂Ǝf[^v邩iACKԂjA
	//	͏Iʒm邩iNACKԂjAw肷B
	TWCR = (1 << TWINT) | (retAck << TWEA) | (1 << TWEN);
	while (!(TWCR & (1 << TWINT)));
	BusErrorCheck();
	
	//Ml擾
	//EM̐ۂɂ炸WX^̒l擾B
	*data = TWDR;
	
	return (Status == (retAck ? STMR_DATA_ACK : STMR_DATA_NACK));	//M̐ۂ͂Ŕ肳
}

//Xgbv𑗐M
void	TWI_Stop()
{
	TWCR = (1 << TWINT) | (1 << TWSTO) | (1 << TWEN);
}
